// git clone https://github.com/vectorclass/version2 vectorclass
// g++ -Ivectorclass -Wall -Wpedantic -Wextra -std=c++17 -march=haswell -O3 sum_array.cpp && ./a.out
#include <random>
#include <algorithm>
//#include <iterator>
#include <iostream>
#include <vector>
#include <numeric>
#include <execution>
#include <cstdint>
#include <cassert>
#include <x86intrin.h>
#include "vectorclass.h"

using namespace std;
#define LENGTH 12800000
#define REPS 10

#define DATA_T float
#define VECTORCLASS_TYPE Vec8f
#define VECTORCLASS_STRIDE 8
#define RANDOM_DISTRIBUTION uniform_real_distribution
//#define DATA_T int32_t
//#define VECTORCLASS_TYPE Vec8i
//#define VECTORCLASS_STRIDE 8
//#define RANDOM_DISTRIBUTION uniform_int_distribution

DATA_T vector_sum_1(vector<DATA_T> vec) {
    DATA_T total = 0;
    for (size_t i = 0; i < vec.size(); i++) {
        total += vec[i];
    }
    return total;
}

DATA_T vector_sum_2(vector<DATA_T> vec) {
    DATA_T total = 0;
    for (auto& elem : vec) {
        total += elem;
    }
    return total;
}

DATA_T vector_sum_3(vector<DATA_T> vec) {
    return std::reduce(vec.cbegin(), vec.cend(), (DATA_T)0.0, [](const auto & x, const auto & y) {
        return x + y;
    });
}

//__attribute__((optimize("-funsafe-math-optimizations")))
DATA_T vector_sum_4(vector<DATA_T> vec) {
    return std::reduce(
        std::execution::unseq,
        vec.cbegin(), vec.cend(),
        (DATA_T)0.0,
        [](const auto & x, const auto & y) {
            return x + y;
        }
    );
}

//__attribute__((optimize("-funsafe-math-optimizations")))
DATA_T array_sum_5(vector<DATA_T> vec) {
    DATA_T total = 0;
    DATA_T* array = &vec[0];
    uint64_t length = vec.size();
    for (size_t i = 0; i < length; i++) {
        total += array[i];
    }
    return total;
}

DATA_T array_sum_6(vector<DATA_T> vec) {
    DATA_T* array = &vec[0];
    uint64_t length = vec.size();
    assert(length % VECTORCLASS_STRIDE == 0);
    VECTORCLASS_TYPE acc = VECTORCLASS_TYPE(0);
    VECTORCLASS_TYPE tmp;

    for (size_t i = 0; i < length; i += VECTORCLASS_STRIDE) {
        tmp.load(array + i);
        acc += tmp;
    }
    return horizontal_add(acc);
}

DATA_T array_sum_7b(DATA_T* array, uint64_t length) {
    DATA_T total = 0;
    for (size_t i = 0; i < length; i++) {
        total += array[i];
    }
    return total;
}

DATA_T array_sum_7(vector<DATA_T> vec) {
    return array_sum_7b(&vec[0], vec.size());
}


void time_vector_sum(string name, DATA_T array_sum(vector<DATA_T>), vector<DATA_T> vec) {
    DATA_T result = 0;
    uint64_t st, en;
    st = __rdtsc();
    for (int i = 0; i < REPS; i++) {
        result += array_sum(vec);
    }
    en = __rdtsc();
    cout << name << " " << result << " in ";
    cout << (double)(en - st) / (double)(vec.size()*REPS) << '\n';
}

int main(void) {
    // https://stackoverflow.com/a/23143753
    random_device rnd_device;
    mt19937 mersenne_engine {rnd_device()};
    RANDOM_DISTRIBUTION<DATA_T> dist {-10, 10};

    auto gen = [&dist, &mersenne_engine]() {
        return dist(mersenne_engine);
    };

    unsigned length = LENGTH;
    vector<DATA_T> vec(length);
    cout << "Array size: " << sizeof(DATA_T) * length / 1024 / 1024 << "MB\n";
    generate(begin(vec), end(vec), gen);

    time_vector_sum("vector_sum_1", vector_sum_1, vec);
    time_vector_sum("vector_sum_2", vector_sum_2, vec);
    time_vector_sum("vector_sum_3", vector_sum_3, vec);
    time_vector_sum("vector_sum_4", vector_sum_4, vec);
    time_vector_sum("array_sum_5 ", array_sum_5, vec);
    time_vector_sum("array_sum_6 ", array_sum_6, vec);
    time_vector_sum("array_sum_7 ", array_sum_7, vec);
    return 0;
}
